#!/usr/bin/python
'''Common routines for converters.

Feb 16 2008
Updated Dec 16 2010'''

import urllib, urllib2, sys, re, htmlentitydefs
import random
from time import time, sleep
from binascii import crc32
import gzip, cStringIO
from hashlib import md5

# Set a default timeout to avoid hung connections
import socket
socket.setdefaulttimeout(30) # seconds

user_agent = 'Mozilla/5.0 (Windows; U; Windows NT 5.1; en-US; rv:1.9.2.13) Gecko/20101203 Firefox/3.6.13'

# redirect stderr: much easier to debug that way
sys.stderr=file('critical_error.log','a')

block_size = 8192
def printline(ln):
    sys.stdout.write('\r'+'%-79s'%ln)
    sys.stdout.flush()

def escape_var(i):
    if isinstance(i,(str,unicode)):
        return i.replace("\\","\\\\").replace("'","\\'")
    return i

def escape_vars(ls):
    return [escape_var(i) for i in ls]

def download_cookies(url, progressline=None, cookiedata='', data=None):
    if isinstance(data, dict):
        data = urllib.urlencode(data)
    if progressline:
        printline(progressline)
    opener = urllib2.OpenerDirector()
    opener.add_handler(urllib2.HTTPHandler())
    req = urllib2.Request(url, headers={'COOKIE':cookiedata, 'User-Agent':user_agent}, data=data)
    conn = opener.open(req)
    cookielist = conn.headers.getheaders('Set-Cookie')
    cookies = []
    for c in cookielist:
        c = c.split('; ', 1)[0]
        cookies.append(c)
    page = conn.read()
    return ('; '.join(cookies), page)

def download_page(url, progressline, cookiedata, retry_limit=10, data=None, sleeptime=1, backoff=1.5):
    for i in range(retry_limit):
        try:
            printline(progressline)
            req=urllib2.Request(url,headers={'COOKIE':cookiedata,'User-Agent':user_agent,'Accept-Encoding':'gzip'}, data=data)
            conn=urllib2.urlopen(req)
            size=int(conn.info().get('Content-Length',0))
            encoding=conn.info().get('Content-Encoding','')
            pagedata=[]
            cursize=0
            block=conn.read(block_size)
            while block:
                pagedata.append(block)
                cursize+=len(block)
                block=conn.read(block_size)
                if size: printline(progressline+"%.1f"%(100.0*cursize/size)+"%")
                else: printline(progressline+"%d bytes"%cursize)
            conn.close()
            if encoding in ('gzip', 'x-gzip'):
                return gzip.GzipFile(fileobj=cStringIO.StringIO(''.join(pagedata)), mode='rb').read()
            return ''.join(pagedata)
        except Exception, detail:
            if retry_limit==1:
                # no retries, so caller probably wants to handle failure
                raise
            elif i==retry_limit-1:
                break # we're done retrying
            printline(progressline+"Failed! "+str(detail)+". Retrying in %d seconds..."%sleeptime)
            print
            sleep(sleeptime)
            # exponential backoff
            sleeptime *= backoff
    printline(progressline+"Failed! Exceeded retry limit - giving up.")
    print
    return ''

def do_except(part, statusline, fn='convert_error.log'):
    olderr=sys.stderr
    sys.stderr=file(fn,'a')
    print >> sys.stderr,part
    sys.excepthook(*sys.exc_info())
    print >> sys.stderr
    printline(statusline+"Exception: See "+fn)
    print
    sys.stderr.close()
    sys.stderr=olderr

def threaded_map(func, iterable, nthreads=None):
    ''' Like map, but with threads. '''
    import threading
    if nthreads is None:
        # automatic
        nthreads = 5
    retval = {}
    control = ['run']
    thread_iter = enumerate(iterable)
    def thread_func():
        while control[0] == 'run':
            try:
                i, val = thread_iter.next()
            except StopIteration:
                return
            retval[i]=func(val)
    try:
        threads = [threading.Thread(target=thread_func, name="threaded_map_%d"%n) for n in xrange(nthreads)]
        for t in threads:
            t.start()
        for t in threads:
            while t.isAlive():
                t.join(1)
    except KeyboardInterrupt:
        control[0] = 'stop'
        print "Caught interrupt signal; exiting threaded_map"
        print "Waiting for threads to finish..."
        for t in threads:
            t.join()
    mx = max(retval)
    ret = []
    for i in xrange(mx+1):
        if i in retval:
            ret.append(retval[i])
        else:
            ret.append(None)
    return ret

def re_pipeline(re_pairs, data, statusline, print_progress=True, debug_prefix=None):
    count = 0
    if debug_prefix is not None:
        file("re_debug_%s_0.txt"%debug_prefix, "w").write(data)
    for search, repl in re_pairs:
        count += 1
        if print_progress:
            printline(statusline+"Processing Step %i of %i"%(count, len(re_pairs)))
        data = re.sub(search, repl, data)
        if debug_prefix is not None:
            file("re_debug_%s_%i.txt"%(debug_prefix, count), "w").write(data)
    return data

# http://effbot.org/zone/re-sub.htm#unescape-html
def re_unescape(m):
    text = m.group(0)
    if text[:2] == "&#":
        # character reference
        try:
            if text[:3] == "&#x":
                return unichr(int(text[3:-1], 16))
            else:
                return unichr(int(text[2:-1]))
        except ValueError:
            pass
    else:
        # named entity
        try:
            text = unichr(htmlentitydefs.name2codepoint[text[1:-1]])
        except KeyError:
            pass
    return text # leave as is

def unescape(text):
    return re.sub("&#?\w+;", re_unescape, text)

def randpw():
    return str(random.randint(1<<32, 1<<40))

def md5hex(data):
    return md5(data).hexdigest()

# -------------------------------------
# Common routines for phpBB3 conversion
# -------------------------------------
def phpbb3_email_hash(email):
    return str(crc32(email)&0xffffffff) + str(len(email))

def phpbb3_makebitfield(dat):
    bbcodes={'code':8,
             'quote':0,
             'attachment':12,
             'b':1,
             'i':2,
             'url':3,
             'img':4,
             'size':5,
             'color':6,
             'u':7,
             'list':9,
             'email':10,
             'flash':11,
             'font':13, # custom BBCodes
             'align':14,
             'spoiler':15,}
    bf=[0]*10
    for i in bbcodes:
        if re.findall('\\[/'+i+':[^\\[\\]]*<UID>]', dat):
            c,d=divmod(bbcodes[i],8)
            bf[c]|=(1<<(7-d))
    return ''.join([chr(c) for c in bf]).rstrip('\0').encode('base64')

def phpbb3_custom_bbcode_sql(table_prefix):
    return 'DELETE FROM '+table_prefix+'bbcodes;INSERT INTO '+table_prefix+r"""bbcodes (bbcode_id, bbcode_tag, bbcode_helpline, display_on_posting, bbcode_match, bbcode_tpl, first_pass_match, first_pass_replace, second_pass_match, second_pass_replace) VALUES
(13, 'font=', 'Font face', 1, '[font={SIMPLETEXT}]{TEXT}[/font]', '<span style="font-family: {SIMPLETEXT};">{TEXT}</span>', '!\\[font\\=([a-zA-Z0-9-+.,_ ]+)\\](.*?)\\[/font\\]!ies', '''[font=${1}:$uid]''.str_replace(array("\\r\\n", ''\\"'', ''\\'''', ''('', '')''), array("\\n", ''"'', ''&#39;'', ''&#40;'', ''&#41;''), trim(''${2}'')).''[/font:$uid]''', '!\\[font\\=([a-zA-Z0-9-+.,_ ]+):$uid\\](.*?)\\[/font:$uid\\]!s', '<span style="font-family: ${1};">${2}</span>'),
(14, 'align=', 'Alignment: can use center, left, right', 1, '[align={SIMPLETEXT}]{TEXT}[/align]', '<div style="text-align: {SIMPLETEXT};">{TEXT}</div>', '!\\[align\\=([a-zA-Z0-9-+.,_ ]+)\\](.*?)\\[/align\\]!ies', '''[align=${1}:$uid]''.str_replace(array("\\r\\n", ''\\"'', ''\\'''', ''('', '')''), array("\\n", ''"'', ''&#39;'', ''&#40;'', ''&#41;''), trim(''${2}'')).''[/align:$uid]''', '!\\[align\\=([a-zA-Z0-9-+.,_ ]+):$uid\\](.*?)\\[/align:$uid\\]!s', '<div style="text-align: ${1};">${2}</div>'),
(15, 'spoiler', 'Spoiler', 1, '[spoiler={TEXT1}]{TEXT2}[/spoiler]', '<dl class="codebox"><dt>{TEXT1}: <a href="#" onclick="if (this.parentNode.parentNode.getElementsByTagName(''dd'')[0].getElementsByTagName(''div'')[0].style.display != '''') { this.parentNode.parentNode.getElementsByTagName(''dd'')[0].getElementsByTagName(''div'')[0].style.display = ''''; this.innerText = ''Hide''; } else { this.parentNode.parentNode.getElementsByTagName(''dd'')[0].getElementsByTagName(''div'')[0].style.display = ''none''; this.innerText = ''Show''; } return false;">Show</a></dt><dd><div style="display: none;">{TEXT2}</div></dd></dl>', '!\\[spoiler(?:\\=(.*?)|)\\](.*?)\\[/spoiler\\]!ies', '''[spoiler=''.(("${1}"=="")?''Spoiler'':str_replace(array("\\r\\n", ''\\"'', ''\\'''', ''('', '')''), array("\\n", ''"'', ''&#39;'', ''&#40;'', ''&#41;''), trim(''${1}''))).'':$uid]''.str_replace(array("\\r\\n", ''\\"'', ''\\'''', ''('', '')''), array("\\n", ''"'', ''&#39;'', ''&#40;'', ''&#41;''), trim(''${2}'')).''[/spoiler:$uid]''', '!\\[spoiler\\=(.*?):$uid\\](.*?)\\[/spoiler:$uid\\]!s', '<dl class="codebox"><dt>${1}: <a href="#" onclick="if (this.parentNode.parentNode.getElementsByTagName(''dd'')[0].getElementsByTagName(''div'')[0].style.display != '''') { this.parentNode.parentNode.getElementsByTagName(''dd'')[0].getElementsByTagName(''div'')[0].style.display = ''''; this.innerText = ''Hide''; } else { this.parentNode.parentNode.getElementsByTagName(''dd'')[0].getElementsByTagName(''div'')[0].style.display = ''none''; this.innerText = ''Show''; } return false;">Show</a></dt><dd><div style="display: none;">${2}</div></dd></dl>');
"""

def phpbb3_default_forum_acl(table_prefix, forumnumber):
    ret=''
    for gid, perm in ((1, 17), # guests: readonly
                      (2, 21), # registered: standard w/ polls
                      (3, 21), # registered+COPPA: standard w/ polls
                      (4, 14), # global mods: full access
                      (4, 11), # global mods: standard moderation
                      (5, 14), # admins: full access
                      (5, 10), # admins: full moderation
                      (6, 19), # bots: bot access
                     ):
        ret+='INSERT INTO '+table_prefix+'acl_groups (group_id, forum_id, auth_option_id, auth_role_id, auth_setting) VALUES (%i, %i, 0, %i, 0);\n'%(gid, forumnumber, perm)
    return ret

confusables = {}
for i in ('.', '..'):
    try:
        f=open(i+"/confusables.dat", 'rb')
    except: continue
    data = f.read().decode('utf8').split('\n')
    for line in data:
        a,b = line.split(':',1)
        confusables[a] = b
    f.close()
    break
else:
    print "Warning: confusables database not found; phpBB3 conversion may be inaccurate!"

def phpbb3_username_clean(username, encoding='utf-8'):
    import unicodedata
    try:
        username=username.decode(encoding)
    except UnicodeDecodeError:
        return username.lower()
    username=username.lower()
    username=unicodedata.normalize('NFKC',username)
    username=u''.join(confusables.get(c, c) for c in username)
    username=re.sub(u'[\x00-\x1f\x7f-\x9f]','',username)
    username=re.sub(' {2,}',' ',username) # reduce multiple spaces to a single one
    return username.strip().encode(encoding)

def phpbb3_bots(table_prefix, first_bot_userid):
    botdata='''		'AdsBot [Google]'			=> array('AdsBot-Google', ''),
		'Alexa [Bot]'				=> array('ia_archiver', ''),
		'Alta Vista [Bot]'			=> array('Scooter/', ''),
		'Ask Jeeves [Bot]'			=> array('Ask Jeeves', ''),
		'Baidu [Spider]'			=> array('Baiduspider+(', ''),
		'Bing [Bot]'                => array('bingbot/', ''),
		'Exabot [Bot]'				=> array('Exabot/', ''),
		'FAST Enterprise [Crawler]'	=> array('FAST Enterprise Crawler', ''),
		'FAST WebCrawler [Crawler]'	=> array('FAST-WebCrawler/', ''),
		'Francis [Bot]'				=> array('http://www.neomo.de/', ''),
		'Gigabot [Bot]'				=> array('Gigabot/', ''),
		'Google Adsense [Bot]'		=> array('Mediapartners-Google', ''),
		'Google Desktop'			=> array('Google Desktop', ''),
		'Google Feedfetcher'		=> array('Feedfetcher-Google', ''),
		'Google [Bot]'				=> array('Googlebot', ''),
		'Heise IT-Markt [Crawler]'	=> array('heise-IT-Markt-Crawler', ''),
		'Heritrix [Crawler]'		=> array('heritrix/1.', ''),
		'IBM Research [Bot]'		=> array('ibm.com/cs/crawler', ''),
		'ICCrawler - ICjobs'		=> array('ICCrawler - ICjobs', ''),
		'ichiro [Crawler]'			=> array('ichiro/', ''),
		'Majestic-12 [Bot]'			=> array('MJ12bot/', ''),
		'Metager [Bot]'				=> array('MetagerBot/', ''),
		'MSN NewsBlogs'				=> array('msnbot-NewsBlogs/', ''),
		'MSN [Bot]'					=> array('msnbot/', ''),
		'MSNbot Media'				=> array('msnbot-media/', ''),
		'NG-Search [Bot]'			=> array('NG-Search/', ''),
		'Nutch [Bot]'				=> array('http://lucene.apache.org/nutch/', ''),
		'Nutch/CVS [Bot]'			=> array('NutchCVS/', ''),
		'OmniExplorer [Bot]'		=> array('OmniExplorer_Bot/', ''),
		'Online link [Validator]'	=> array('online link validator', ''),
		'psbot [Picsearch]'			=> array('psbot/0', ''),
		'Seekport [Bot]'			=> array('Seekbot/', ''),
		'Sensis [Crawler]'			=> array('Sensis Web Crawler', ''),
		'SEO Crawler'				=> array('SEO search Crawler/', ''),
		'Seoma [Crawler]'			=> array('Seoma [SEO Crawler]', ''),
		'SEOSearch [Crawler]'		=> array('SEOsearch/', ''),
		'Snappy [Bot]'				=> array('Snappy/1.1 ( http://www.urltrends.com/ )', ''),
		'Steeler [Crawler]'			=> array('http://www.tkl.iis.u-tokyo.ac.jp/~crawler/', ''),
		'Synoo [Bot]'				=> array('SynooBot/', ''),
		'Telekom [Bot]'				=> array('crawleradmin.t-info@telekom.de', ''),
		'TurnitinBot [Bot]'			=> array('TurnitinBot/', ''),
		'Voyager [Bot]'				=> array('voyager/1.0', ''),
		'W3 [Sitesearch]'			=> array('W3 SiteSearch Crawler', ''),
		'W3C [Linkcheck]'			=> array('W3C-checklink/', ''),
		'W3C [Validator]'			=> array('W3C_*Validator', ''),
		'WiseNut [Bot]'				=> array('http://www.WISEnutbot.com', ''),
		'YaCy [Bot]'				=> array('yacybot', ''),
		'Yahoo MMCrawler [Bot]'		=> array('Yahoo-MMCrawler/', ''),
		'Yahoo Slurp [Bot]'			=> array('Yahoo! DE Slurp', ''),
		'Yahoo [Bot]'				=> array('Yahoo! Slurp', ''),
		'YahooSeeker [Bot]'			=> array('YahooSeeker/', ''),'''

    bot_list=[]
    for line in botdata.split('\n'):
        name, arr=line.strip().strip(',').split('=>',1)
        bot_list.append((eval(name), eval(arr, {'array':lambda *args:args})))

    ret = ["DELETE FROM "+table_prefix+'bots;\n']
    for i, (bot_name, bot_arr) in enumerate(bot_list):
        userid = first_bot_userid+i
        ret.append("INSERT INTO "+table_prefix+"users (user_id, user_type, group_id, username, username_clean, user_regdate, user_password, user_colour, user_email, user_email_hash, user_lang, user_style, user_timezone, user_dateformat, user_allow_massemail) VALUES ")
        ret.append("(%i, 2, 6, '%s', '%s', %i, '', '9E8DA7', '', '00', 'english', 1, 0, 'D M d, Y g:i a', 0);\n"%(userid, bot_name, bot_name.lower(), time()))
        ret.append("INSERT INTO "+table_prefix+"user_group (group_id, user_id, user_pending, group_leader) VALUES (6, %i, 0, 0);\n"%userid)
        ret.append("INSERT INTO "+table_prefix+"bots (bot_active, bot_name, user_id, bot_agent, bot_ip) VALUES (1, '%s', %i, '%s', '%s');\n"%(bot_name, userid, bot_arr[0], bot_arr[1]))

    return ''.join(ret)
